#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NUM_TRACKS	(35)
#define NUM_SECTORS	(16)
#define SECTOR_SIZE	(0x1A0)
#define TRACK_SIZE	(SECTOR_SIZE * NUM_SECTORS)
#define GAP_THRETHOLD	(16)

struct trk_sec_t {
	int trk;
	int sec;
};


static void usage(void);
static int get_long_param(char *str, long *value);
static int get_uchar_param(char *str, unsigned char *value);
static int get_trk_sec_param(char *str, struct trk_sec_t *value);
static int read_file(const char *filename, unsigned char **buffer, long *size);
static int dump_nib_image(const unsigned char *buffer, long size, long offset, long length, const struct trk_sec_t *start_ts, const struct trk_sec_t *end_ts);
static int search_header(const unsigned char *buffer, int size, int offset);
static int search_data(const unsigned char *buffer, int size, int offset);
static unsigned char get_next_byte(const unsigned char *buffer, int size, int offset, int *next);
static void setup_trans_table(void);
static int decode_sector_data(unsigned char *decoded_data, const unsigned char *sector_data, unsigned char *check_sum);


static int opt_decode_sector_data = 0;
static long gap_threthold = GAP_THRETHOLD;
static unsigned char trans_value_AA = 0xFF;
static unsigned char trans_value_D5 = 0xFF;

static const unsigned char write_trans_table_6_and_2[64] = {
	0x96, 0x97, 0x9A, 0x9B, 0x9D, 0x9E, 0x9F, 0xA6, 0xA7, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF, 0xB2, 0xB3,
	0xB4, 0xB5, 0xB6, 0xB7, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF, 0xCB, 0xCD, 0xCE, 0xCF, 0xD3,
	0xD6, 0xD7, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF, 0xE5, 0xE6, 0xE7, 0xE9, 0xEA, 0xEB, 0xEC,
	0xED, 0xEE, 0xEF, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
};

static const int sector_skewing_table[16] = {
	0, 7, 14, 6, 13, 5, 12, 4, 11, 3, 10, 2, 9, 1, 8, 15	/* DOS 3.3 */
};

static unsigned char read_trans_table_6_and_2[256];


int main(int argc, char **argv)
{
	int result = 1;
	int num_cmd_params = 1;
	const char *image_file = NULL;
	unsigned char *image_ptr = NULL;
	long image_size = 0;
	long offset = 0;
	long length = 0;
	struct trk_sec_t start_trk_sec = { 0, 0 };
	struct trk_sec_t end_trk_sec = { -1, -1 };

	setup_trans_table();

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0 ||
				strcmp(*argv, "-?") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--offset") == 0 ||
					   strcmp(*argv, "-o") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no offset\n");
					return 1;
				}
				if (get_long_param(*argv, &offset) < 0) {
					printf("ERROR: illegal offset '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--length") == 0 ||
					   strcmp(*argv, "-l") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no length\n");
					return 1;
				}
				if (get_long_param(*argv, &length) < 0) {
					printf("ERROR: illegal length '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--start") == 0 ||
					   strcmp(*argv, "-s") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no trk/sec\n");
					return 1;
				}
				if (get_trk_sec_param(*argv, &start_trk_sec) < 0) {
					printf("ERROR: illegal trk/sec '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--end") == 0 ||
					   strcmp(*argv, "-e") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no trk/sec\n");
					return 1;
				}
				if (get_trk_sec_param(*argv, &end_trk_sec) < 0) {
					printf("ERROR: illegal trk/sec '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--gap") == 0 ||
					   strcmp(*argv, "-g") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no gap threthold\n");
					return 1;
				}
				if (get_long_param(*argv, &gap_threthold) < 0) {
					printf("ERROR: illegal gap threthold '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--decode") == 0 ||
					   strcmp(*argv, "-d") == 0) {
				opt_decode_sector_data = 1;
			} else if (strcmp(*argv, "--AA") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no $AA value\n");
					return 1;
				}
				if (get_uchar_param(*argv, &trans_value_AA) < 0 || 64 <= trans_value_AA) {
					printf("ERROR: illegal $AA value '%s'\n", *argv);
					return 1;
				}
			} else if (strcmp(*argv, "--D5") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: no $D5 value\n");
					return 1;
				}
				if (get_uchar_param(*argv, &trans_value_D5) < 0 || 64 <= trans_value_D5) {
					printf("ERROR: illegal $D5 value '%s'\n", *argv);
					return 1;
				}
			} else {
				printf("ERROR: unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("ERROR: illegal parameter '%s'\n", *argv);
				return 1;
			}
			image_file = *argv;
			num_cmd_params--;
		}
	}

	if (image_file == NULL) {
		printf("ERROR: no image file\n");
		goto ERROR_EXIT;
	}

	if (read_file(image_file, &image_ptr, &image_size) < 0) {
		printf("ERROR: read error '%s'\n", image_file);
		goto ERROR_EXIT;
	}

	if (dump_nib_image(image_ptr, image_size, offset, length, &start_trk_sec, &end_trk_sec) < 0) {
		printf("ERROR: dump error '%s'\n", image_file);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:

	return result;
}

static void usage(void)
{
	printf("usage: nib_dump [options] image.nib\n");
	printf("option: -o|--offset         specify offset of disk image\n");
	printf("        -l|--length         specify length of disk image\n");
	printf("        -s|--start TRK/SEC  specify start track/sector\n");
	printf("        -e|--end   TRK/SEC  specify end track/sector\n");
	printf("        -g|--gap THRETHOLD  specify gap threthold (default=%d)\n", GAP_THRETHOLD);
	printf("        -d|--decode         decode sector data\n");
	printf("           --AA    VALUE    translate $AA to VALUE when decoding\n");
	printf("           --D5    VALUE    translate $D5 to VALUE when decoding\n");
	printf("                            VALUE = 0 ..63\n");
	printf("        -?|--help           print usage\n");
}

static int get_long_param(char *str, long *value)
{
	char *endptr;

	if (value == NULL) {
		printf("ERROR: BUG: pointer is NULL\n");
		return -1;
	}

	*value = strtoul(str, &endptr, 0);
	if (endptr == NULL || *endptr != '\0') {
		return -1;
	}

	return 0;
}

static int get_uchar_param(char *str, unsigned char *value)
{
	char *endptr;
	unsigned long work;

	if (value == NULL) {
		printf("ERROR: BUG: pointer is NULL\n");
		return -1;
	}

	work = strtoul(str, &endptr, 0);
	if (endptr == NULL || *endptr != '\0' || 256 <= work) {
		return -1;
	}

	*value = (unsigned char)work;

	return 0;
}

static int get_trk_sec_param(char *str, struct trk_sec_t *value)
{
	char *endptr;

	if (value == NULL) {
		printf("ERROR: BUG: pointer is NULL\n");
		return -1;
	}

	value->trk = strtoul(str, &endptr, 0);
	if (endptr == NULL || *endptr != '/' || NUM_TRACKS <= value->trk) {
		return -1;
	}

	value->sec = strtoul(endptr + 1, &endptr, 0);
	if (endptr == NULL || *endptr != '\0' || NUM_SECTORS <= value->sec) {
		return -1;
	}

	return 0;
}

static int read_file(const char *filename, unsigned char **buffer, long *size)
{
	int result = -1;
	FILE *fp = NULL;

	if (buffer == NULL || size == NULL) {
		printf("ERROR: BUG: pointer is NULL\n");
		goto ERROR_EXIT;
	}

	*buffer = NULL;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: can't open file '%s'\n", filename);
		goto ERROR_EXIT;
	}

	if (fseek(fp, 0, SEEK_END) < 0) {
		printf("ERROR: can't seek to end '%s'\n", filename);
		goto ERROR_EXIT;
	}
	(*size) = ftell(fp);
	if (fseek(fp, 0, SEEK_SET) < 0) {
		printf("ERROR: can't seek to head '%s'\n", filename);
		goto ERROR_EXIT;
	}

	if (((*buffer) = (unsigned char *)malloc(*size)) == NULL) {
		printf("ERROR: memory exhausted: size=%ld\n", *size);
		goto ERROR_EXIT;
	}

	if (fread(*buffer, *size, 1, fp) != 1) {
		printf("ERROR: can't read file '%s'\n", filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int dump_nib_image(const unsigned char *buffer, long size, long offset, long length, const struct trk_sec_t *start_ts, const struct trk_sec_t *end_ts)
{
	int trk;
	int trk_offset;
	int trk_size;

	if (buffer == NULL) {
		printf("ERROR: BUG: pointer is NULL\n");
		return -1;
	}

	if (size <= offset || (0 < length && size <= (offset + length))) {
		printf("ERROR: illegal offset/length: offset=%ld, length=%ld, size=%ld\n", offset, length, size);
		return -1;
	}

	if (length <= 0) {
		length = size - offset;
	}

	buffer += offset;
	for (trk = start_ts->trk; end_ts->trk < 0 || trk <= end_ts->trk; trk++) {
		int trk_start;
		int offs;

		trk_offset = trk * TRACK_SIZE;
		if (length <= trk_offset) {
			break;
		}

		trk_size = TRACK_SIZE;
		if (length < trk_offset + trk_size) {
			trk_size = length - trk_offset;
			printf("WARNING: offset=%08X, track=%02X(%d), size=%08X(%d)\n", trk_offset, trk, trk, trk_size, trk_size);
		}

		if ((trk_start = search_header(buffer + trk_offset, trk_size, 0)) < 0) {
			printf("WARNING: can't find track: offset=%08X, track=%02X(%d)\n", trk_offset, trk, trk);
			continue;
		}

		printf("------------------------------------------------------------------------------------\n");
		printf("TRACK: %02X(%d) %08X - %08X\n", trk, trk, trk_offset, trk_offset + trk_size - 1);
		printf("------------------------------------------------------------------------------------\n");

		offs = trk_start;
		while (1) {
			unsigned char header[14];
			unsigned char hdr_vol;
			unsigned char hdr_trk;
			unsigned char hdr_sec;
			unsigned char hdr_sum;
			unsigned char data_end[3];
			int save;
			int rest;
			const unsigned char end_mark[3] = { 0xDE, 0xAA, 0xEB };
			int i;

			if (trk_start <= offs) {
				rest = trk_size - offs + trk_start;
			} else {
				rest = trk_start - offs;
			}
			if (rest < (3 + 8 + 3 + 3 + 0x157 + 3)) {
				printf("WARNING: garbage ?: track=%02X(%d), offset=%08X + %04X:\n", trk, trk, trk_offset, offs);

				save = offs;
				offs = trk_start;

				printf("UNKNOWN:\n");
				while (save != offs) {
					printf("%08X:", trk_offset + save);
					for (i = 0; i < 16 && save != offs; i++) {
						printf(" %02X", get_next_byte(buffer + trk_offset, trk_size, save, &save));
					}
					printf("\n");
				}
				printf("\n");

				break;
			}

			save = offs;
			for (i = 0; i < 14; i++) {
				header[i] = get_next_byte(buffer + trk_offset, trk_size, offs, &offs);
			}
			hdr_vol = ((header[3] & 0x55) << 1) | (header[4] & 0x55);
			hdr_trk = ((header[5] & 0x55) << 1) | (header[6] & 0x55);
			hdr_sec = ((header[7] & 0x55) << 1) | (header[8] & 0x55);
			hdr_sum = ((header[9] & 0x55) << 1) | (header[10] & 0x55);

			if ((start_ts->trk < trk || (start_ts->trk == trk && start_ts->sec <= hdr_sec)) &&
				(end_ts->trk < 0 || trk < end_ts->trk || (end_ts->trk == trk && hdr_sec <= end_ts->sec))) {
				printf("ADDRESS:\n");
				printf("%08X:", trk_offset + save);
				for (i = 0; i < 14; i++) {
					printf(" %02X", header[i]);
				}
				printf(" : vol=%02X, trk=%02X, sec=%02X, sum=%02X", hdr_vol, hdr_trk, hdr_sec, hdr_sum);
				if (header[11] != end_mark[0] ||
					header[12] != end_mark[1] ||
					header[13] != end_mark[2]) {
					printf(" : check end mark");
				}
				printf("\n");

				save = offs;
				if ((offs = search_data(buffer + trk_offset, trk_size, offs)) < 0) {
					printf("WARNING: data not found\n");
					offs = save;
				} else {
					int gap_size = 0;
					unsigned char sector_data[0x157];

					printf("GAP:\n");
					while (save != offs) {
						printf("%08X:", trk_offset + save);
						for (i = 0; i < 16 && save != offs; i++, gap_size++) {
							printf(" %02X", get_next_byte(buffer + trk_offset, trk_size, save, &save));
						}
						printf("\n");
					}
					if (gap_threthold < gap_size) {
						printf("WARNING: too large gap!: \n");
					}

					printf("DATA:\n");
					printf("%08X:", trk_offset + offs);
					for (i = 0; i < 3; i++) {
						printf(" %02X", get_next_byte(buffer + trk_offset, trk_size, offs, &offs));
					}
					printf("\n");

					for (i = 0; i < 0x157; ) {
						int j;
						printf("%08X:", trk_offset + offs);
						for (j = 0; j < 16 && i < 0x157; j++, i++) {
							sector_data[i] = get_next_byte(buffer + trk_offset, trk_size, offs, &offs);
							printf("%c%02X", (read_trans_table_6_and_2[sector_data[i]] < 64 ? ' ' : '>'), sector_data[i]);
						}
						printf("\n");
					}

					printf("%08X:", trk_offset + offs);
					for (i = 0; i < 3; i++) {
						data_end[i] = get_next_byte(buffer + trk_offset, trk_size, offs, &offs);
						printf(" %02X", data_end[i]);
					}
					if (data_end[0] != end_mark[0] ||
						data_end[1] != end_mark[1] ||
						data_end[2] != end_mark[2]) {
						printf(" : check end mark\n");
					}
					printf("\n");

					if (opt_decode_sector_data) {
						unsigned char check_sum;
						unsigned char decoded_data[256];

						if (decode_sector_data(decoded_data, sector_data, &check_sum) < 0) {
							printf("WARNING: can't decode sector data\n");
						} else {
							if (check_sum != 0) {
								printf("WARNING: check sum error: %02X\n", check_sum);
							}
							printf("DECODED:");
							if (hdr_sec < 16) {
								printf(" sector skewing: %X (physical) -> %X (DOS 3.3)", hdr_sec, sector_skewing_table[hdr_sec]);
							}
							printf("\n");
							for (i = 0; i < 256; ) {
								int j;
								printf("   ");
								for (j = 0; j < 16; j++, i++) {
									printf(" %02X", decoded_data[i]);
								}
								printf("\n");
							}
						}
					}
				}

				save = offs;
				if ((offs = search_header(buffer + trk_offset, trk_size, offs)) < 0) {
					offs = trk_start;
				}

				printf("GAP:\n");
				while (save != offs) {
					printf("%08X:", trk_offset + save);
					for (i = 0; i < 16 && save != offs; i++) {
						printf(" %02X", get_next_byte(buffer + trk_offset, trk_size, save, &save));
					}
					printf("\n");
				}
				printf("\n");

			} else {
				/* skip */
				if ((offs = search_header(buffer + trk_offset, trk_size, offs)) < 0) {
					/* not found */
					break;
				}
			}
			if (offs == trk_start) {
				/* end */
				break;
			}
		}

	}

	return 0;
}

static int search_header(const unsigned char *buffer, int size, int offset)
{
	int i;

	for (i = 0; i < size; i++) {
		int save = offset;
		int tmp;
		if (get_next_byte(buffer, size, offset, &offset) == 0xD5 &&
			get_next_byte(buffer, size, offset, &tmp   ) == 0xAA &&
			get_next_byte(buffer, size, tmp   , &tmp   ) == 0x96) {
			return save;
		}
	}

	return -1;
}

static int search_data(const unsigned char *buffer, int size, int offset)
{
	int i;

	for (i = 0; i < size; i++) {
		int save = offset;
		int tmp;
		if (get_next_byte(buffer, size, offset, &offset) == 0xD5 &&
			get_next_byte(buffer, size, offset, &tmp   ) == 0xAA) {
			unsigned char value = get_next_byte(buffer, size, tmp, &tmp);
			if (value == 0xAD) {
				return save;
			}
			if (value == 0x96) {
				return -1;
			}
		}
	}

	return -1;
}

static unsigned char get_next_byte(const unsigned char *buffer, int size, int offset, int *next)
{
	unsigned char value;
	value = buffer[offset++];
	if (size <= offset) {
		offset = 0;
	}
	*next = offset;
	return value;
}

static void setup_trans_table(void)
{
	int i;

	memset(read_trans_table_6_and_2, 0xFF, sizeof(read_trans_table_6_and_2));

	for (i = 0; i < 64; i++) {
		int index = write_trans_table_6_and_2[i];
		if (read_trans_table_6_and_2[index] != 0xFF) {
			printf("ERROR: BUG: setup_trans_table(): i=%d, index=%d\n", i, index);
			exit(1);
		}
		read_trans_table_6_and_2[index] = i;
	}
}

static int decode_sector_data(unsigned char *decoded_data, const unsigned char *sector_data, unsigned char *check_sum)
{
	int i;
	unsigned char translated[0x157];
	unsigned char sum = 0;

	memset(decoded_data, 0, sizeof(decoded_data));

	for (i = 0; i < 0x157; i++) {
		unsigned char value = read_trans_table_6_and_2[sector_data[i]];
		if (64 <= value) {
			printf("WARNING: illegal sector data: %02X, offs=%04X, sum=%02X\n", sector_data[i], i, sum);
			if (sector_data[i] == 0xAA && trans_value_AA < 64) {
				value = trans_value_AA;
			} else if (sector_data[i] == 0xD5 && trans_value_D5 < 64) {
				value = trans_value_D5;
			} else {
				return -1;
			}
		}
		translated[i] = value ^ sum;
		sum = translated[i];
	}

	*check_sum = sum;

	for (i = 0; i < 0x56; i++) {
		const unsigned char reverse_bits[4] = { 0, 2, 1, 3 };
		decoded_data[i] = reverse_bits[translated[i] & 0x3];
		decoded_data[i + 0x56] = reverse_bits[(translated[i] >> 2) & 0x3];
		if (i + 0xAC < 256) {
			decoded_data[i + 0xAC] = reverse_bits[(translated[i] >> 4) & 0x3];
		}
	}

	for (i = 0; i < 256; i++) {
		decoded_data[i] |= (translated[i + 0x56] << 2);
	}

	return 0;
}

/* EOF */
